#include "hbhi.h"
#include "launcher/launcher.h"
#include "rawinput/rawinput.h"
#include "misc/eamuse.h"
#include "games/rf3d/io.h"
#include "games/sc/io.h"
#include "avs/game.h"

using namespace GameAPI;

/*
 * Implementations
 */

static int __cdecl ac_io_hbhi_add_coin(int a1, int a2) {
    eamuse_coin_add();
    return 1;
}

static char __cdecl ac_io_hbhi_consume_coinstock(int a1, int a2) {
    eamuse_coin_consume_stock();
    return 1;
}

static int __cdecl ac_io_hbhi_control_coin_blocker_close(int a1) {
    eamuse_coin_set_block(true);
    return 1;
}

static int __cdecl ac_io_hbhi_control_coin_blocker_open(int a1) {
    eamuse_coin_set_block(0);
    return 1;
}

static int __cdecl ac_io_hbhi_control_lamp_bright(int a1, unsigned __int8 a2) {
    return 1;
}

static int __cdecl ac_io_hbhi_control_lamp_mode(int a1) {
    return 1;
}

static int __cdecl ac_io_hbhi_control_lamp_off(int a1) {
    return 1;
}

static int __cdecl ac_io_hbhi_control_lamp_on(char a1) {
    return 1;
}

static int __cdecl ac_io_hbhi_control_parallel_off(char a1) {
    return 1;
}

static int __cdecl ac_io_hbhi_control_parallel_on(char a1) {
    return 1;
}

static int __cdecl ac_io_hbhi_control_reset() {
    return 1;
}

static bool __cdecl ac_io_hbhi_create_get_status_thread(void *a1) {
    return true;
}

static char __cdecl ac_io_hbhi_current_coinstock(int a1, int *a2) {
    *a2 = eamuse_coin_get_stock();
    return 1;
}

static int __cdecl ac_io_hbhi_destroy_get_status_thread() {
    return 1;
}

static char __cdecl ac_io_hbhi_get_coin_input_wave_buffer(int *a1) {
    return 1;
}

static void *__cdecl ac_io_hbhi_get_control_status_buffer(void *a1) {

    // steel chronicles
    if (avs::game::is_model("KGG")) {

        // get buttons
        auto buttons = games::sc::get_buttons();

        // reset
        memset(a1, 0x00, 64);

        // check buttons
        if (Buttons::getState(RI_MGR, buttons->at(games::sc::Buttons::Service)))
            *((char *) a1 + 5) |= 1 << 4;
        if (Buttons::getState(RI_MGR, buttons->at(games::sc::Buttons::Test)))
            *((char *) a1 + 5) |= 1 << 5;
    }

    // road fighters 3D
    if (avs::game::is_model("JGT")) {
        static int lever_state = 0;

        // get buttons
        auto buttons = games::rf3d::get_buttons();

        // reset
        memset(a1, 0x00, 64);

        // check buttons
        if (Buttons::getState(RI_MGR, buttons->at(games::rf3d::Buttons::Service)))
            *((char *) a1 + 5) |= 1 << 4;
        if (Buttons::getState(RI_MGR, buttons->at(games::rf3d::Buttons::Test)))
            *((char *) a1 + 5) |= 1 << 5;
        if (Buttons::getState(RI_MGR, buttons->at(games::rf3d::Buttons::CoinMech)))
            *((char *) a1 + 5) |= 1 << 2;
        if (Buttons::getState(RI_MGR, buttons->at(games::rf3d::Buttons::View)))
            *((char *) a1 + 12) |= 1 << 2;
        if (Buttons::getState(RI_MGR, buttons->at(games::rf3d::Buttons::Toggle2D3D)))
            *((char *) a1 + 12) |= 1 << 3;
        if (Buttons::getState(RI_MGR, buttons->at(games::rf3d::Buttons::LeverUp))) {
            *((char *) a1 + 12) |= 1 << 4;
            lever_state = 0;
        }
        if (Buttons::getState(RI_MGR, buttons->at(games::rf3d::Buttons::LeverDown))) {
            *((char *) a1 + 12) |= 1 << 5;
            lever_state = 0;
        }
        if (Buttons::getState(RI_MGR, buttons->at(games::rf3d::Buttons::LeverLeft))) {
            *((char *) a1 + 12) |= 1 << 6;
            lever_state = 0;
        }
        if (Buttons::getState(RI_MGR, buttons->at(games::rf3d::Buttons::LeverRight))) {
            *((char *) a1 + 12) |= 1 << 7;
            lever_state = 0;
        }

        // auto lever buttons
        if (Buttons::getState(RI_MGR, buttons->at(games::rf3d::Buttons::AutoLeverUp)) && lever_state < 6)
            lever_state++;
        if (Buttons::getState(RI_MGR, buttons->at(games::rf3d::Buttons::AutoLeverDown)) && lever_state > 0)
            lever_state--;

        // auto lever logic
        switch (lever_state) {
            case 1:
                *((char *) a1 + 12) |= 1 << 4 | 1 << 6;
                break;
            case 2:
                *((char *) a1 + 12) |= 1 << 4 | 1 << 6;
                break;
            case 3:
                *((char *) a1 + 12) |= 1 << 4;
                break;
            case 4:
                *((char *) a1 + 12) |= 1 << 5;
                break;
            case 5:
                *((char *) a1 + 12) |= 1 << 4 | 1 << 7;
                break;
            case 6:
                *((char *) a1 + 12) |= 1 << 5 | 1 << 7;
                break;
            default:
                lever_state = 0;
                break;
        }
    }

    // return buffer
    return a1;
}

static char __cdecl ac_io_hbhi_get_softwareid(char *a1) {
    memset(a1, 'F', 16);
    return 1;
}

static char __cdecl ac_io_hbhi_get_systemid(char *a1) {
    memset(a1, 'F', 16);
    return 1;
}

static bool __cdecl ac_io_hbhi_get_watchdog_status() {
    return true;
}

static short __cdecl ac_io_hbhi_get_watchdog_time_min() {
    return 0;
}

static short __cdecl ac_io_hbhi_get_watchdog_time_now() {
    return 0;
}

static char __cdecl ac_io_hbhi_lock_coincounter(int a1) {
    return 1;
}

static char __cdecl ac_io_hbhi_req_carddispenser_disburse() {
    return 1;
}

static bool __cdecl ac_io_hbhi_req_carddispenser_disburse_isfinished(int *a1) {
    *a1 += 1;
    return true;
}

static char __cdecl ac_io_hbhi_req_carddispenser_get_status() {
    return 1;
}

static int __cdecl ac_io_hbhi_req_carddispenser_get_status_isfinished(int *a1) {
    *a1 += 1;
    return 2;
}

static char __cdecl ac_io_hbhi_req_carddispenser_init() {
    return 1;
}

static bool __cdecl ac_io_hbhi_req_carddispenser_init_isfinished(int *a1) {
    *a1 += 1;
    return true;
}

static char __cdecl ac_io_hbhi_req_coin_input_wave() {
    return 1;
}

static char __cdecl ac_io_hbhi_req_get_control_status(int *a1) {
    return 1;
}

static char __cdecl ac_io_hbhi_req_secplug_check(char *a1) {
    return 1;
}

static bool __cdecl ac_io_hbhi_req_secplug_check_isfinished(int *a1) {
    return true;
}

static char __cdecl ac_io_hbhi_req_secplug_check_softwareplug(char *a1) {
    return 1;
}

static char __cdecl ac_io_hbhi_req_secplug_check_systemplug() {
    return 1;
}

static char __cdecl ac_io_hbhi_req_secplug_missing_check() {
    return 1;
}

static bool __cdecl ac_io_hbhi_req_secplug_missing_check_isfinished(int *a1) {
    return true;
}

static bool __cdecl ac_io_hbhi_req_volume_control(char a1, char a2) {
    return true;
}

static bool __cdecl ac_io_hbhi_req_volume_control_isfinished(int *a1) {
    return true;
}

static int __cdecl ac_io_hbhi_reset_coin_slot_noise_flag(int a1) {
    return 1;
}

static int __cdecl ac_io_hbhi_set_framing_err_packet_send_interval(int a1) {
    return 1;
}

static bool __cdecl ac_io_hbhi_set_watchdog_time(short a1) {
    return true;
}

static char __cdecl ac_io_hbhi_unlock_coincounter(int a1) {
    return 1;
}

static char __cdecl ac_io_hbhi_update_control_status_buffer() {
    return 1;
}

static void __cdecl ac_io_hbhi_watchdog_off() {
}

/*
 * Module stuff
 */

acio::HBHIModule::HBHIModule(HMODULE module, acio::HookMode hookMode) : ACIOModule("HBHI", module, hookMode) {
}

void acio::HBHIModule::attach() {
    ACIOModule::attach();

    // hooks
    ACIO_MODULE_HOOK(ac_io_hbhi_add_coin);
    ACIO_MODULE_HOOK(ac_io_hbhi_consume_coinstock);
    ACIO_MODULE_HOOK(ac_io_hbhi_control_coin_blocker_close);
    ACIO_MODULE_HOOK(ac_io_hbhi_control_coin_blocker_open);
    ACIO_MODULE_HOOK(ac_io_hbhi_control_lamp_bright);
    ACIO_MODULE_HOOK(ac_io_hbhi_control_lamp_mode);
    ACIO_MODULE_HOOK(ac_io_hbhi_control_lamp_off);
    ACIO_MODULE_HOOK(ac_io_hbhi_control_lamp_on);
    ACIO_MODULE_HOOK(ac_io_hbhi_control_parallel_off);
    ACIO_MODULE_HOOK(ac_io_hbhi_control_parallel_on);
    ACIO_MODULE_HOOK(ac_io_hbhi_control_reset);
    ACIO_MODULE_HOOK(ac_io_hbhi_create_get_status_thread);
    ACIO_MODULE_HOOK(ac_io_hbhi_current_coinstock);
    ACIO_MODULE_HOOK(ac_io_hbhi_destroy_get_status_thread);
    ACIO_MODULE_HOOK(ac_io_hbhi_get_coin_input_wave_buffer);
    ACIO_MODULE_HOOK(ac_io_hbhi_get_control_status_buffer);
    ACIO_MODULE_HOOK(ac_io_hbhi_get_softwareid);
    ACIO_MODULE_HOOK(ac_io_hbhi_get_systemid);
    ACIO_MODULE_HOOK(ac_io_hbhi_get_watchdog_status);
    ACIO_MODULE_HOOK(ac_io_hbhi_get_watchdog_time_min);
    ACIO_MODULE_HOOK(ac_io_hbhi_get_watchdog_time_now);
    ACIO_MODULE_HOOK(ac_io_hbhi_lock_coincounter);
    ACIO_MODULE_HOOK(ac_io_hbhi_req_carddispenser_disburse);
    ACIO_MODULE_HOOK(ac_io_hbhi_req_carddispenser_disburse_isfinished);
    ACIO_MODULE_HOOK(ac_io_hbhi_req_carddispenser_get_status);
    ACIO_MODULE_HOOK(ac_io_hbhi_req_carddispenser_get_status_isfinished);
    ACIO_MODULE_HOOK(ac_io_hbhi_req_carddispenser_init);
    ACIO_MODULE_HOOK(ac_io_hbhi_req_carddispenser_init_isfinished);
    ACIO_MODULE_HOOK(ac_io_hbhi_req_coin_input_wave);
    ACIO_MODULE_HOOK(ac_io_hbhi_req_get_control_status);
    ACIO_MODULE_HOOK(ac_io_hbhi_req_secplug_check);
    ACIO_MODULE_HOOK(ac_io_hbhi_req_secplug_check_isfinished);
    ACIO_MODULE_HOOK(ac_io_hbhi_req_secplug_check_softwareplug);
    ACIO_MODULE_HOOK(ac_io_hbhi_req_secplug_check_systemplug);
    ACIO_MODULE_HOOK(ac_io_hbhi_req_secplug_missing_check);
    ACIO_MODULE_HOOK(ac_io_hbhi_req_secplug_missing_check_isfinished);
    ACIO_MODULE_HOOK(ac_io_hbhi_req_volume_control);
    ACIO_MODULE_HOOK(ac_io_hbhi_req_volume_control_isfinished);
    ACIO_MODULE_HOOK(ac_io_hbhi_reset_coin_slot_noise_flag);
    ACIO_MODULE_HOOK(ac_io_hbhi_set_framing_err_packet_send_interval);
    ACIO_MODULE_HOOK(ac_io_hbhi_set_watchdog_time);
    ACIO_MODULE_HOOK(ac_io_hbhi_unlock_coincounter);
    ACIO_MODULE_HOOK(ac_io_hbhi_update_control_status_buffer);
    ACIO_MODULE_HOOK(ac_io_hbhi_watchdog_off);
}
