#include "foot.h"
#include "util/logging.h"
#include "rawinput/rawinput.h"
#include "io.h"

bool games::ddr::DDRFOOTHandle::open(LPCWSTR lpFileName) {
    if (wcscmp(lpFileName, L"COM1") != 0)
        return false;
    log_info("ddr", "Opened COM1 (FOOT).");
    return true;
}

int games::ddr::DDRFOOTHandle::read(LPVOID lpBuffer, DWORD nNumberOfBytesToRead) {

    // check data trigger and buffer size
    if (data_trigger && nNumberOfBytesToRead >= 1) {
        data_trigger = false;
        memset(lpBuffer, 0x11, 1);
        return 1;
    }

    // no data
    return 0;
}

int games::ddr::DDRFOOTHandle::write(LPCVOID lpBuffer, DWORD nNumberOfBytesToWrite) {

    // check buffer size for lights
    if (nNumberOfBytesToWrite == 4) {

        // mappings
        static const size_t mapping_bits[] = {
                4, 6, 3, 5,     // P1 L, U, R, D,
                12, 14, 11, 13, // P2 L, U, R, D
                22              // NEON
        };
        
        static const size_t mapping[] = {
                Lights::P1FootLeft,
                Lights::P1FootUp,
                Lights::P1FootRight,
                Lights::P1FootDown,
                Lights::P2FootLeft,
                Lights::P2FootUp,
                Lights::P2FootRight,
                Lights::P2FootDown,
                Lights::Neon
        };

        // get light bits
        uint32_t light_bits = *((uint32_t*) lpBuffer);

        // get lights
        auto lights = get_lights();

        // bit scan
        for (size_t i = 0; i < 9; i++) {
            float value = (light_bits & (1 << mapping_bits[i])) ? 1.f : 0.f;
            GameAPI::Lights::writeLight(RI_MGR, lights->at(mapping[i]), value);
        }     

        // flush
        RI_MGR->devices_flush_output();
    }

    // trigger out data
    data_trigger = true;

    // return all data written
    return (int) nNumberOfBytesToWrite;
}

int games::ddr::DDRFOOTHandle::device_io(DWORD dwIoControlCode, LPVOID lpInBuffer, DWORD nInBufferSize, LPVOID lpOutBuffer,
                                     DWORD nOutBufferSize) {
    return -1;
}

bool games::ddr::DDRFOOTHandle::close() {
    log_info("ddr", "Closed COM1 (FOOT).");
    return true;
}
