#include "setupapihook.h"
#include <windows.h>
#include <setupapi.h>
#include "util/detour.h"
#include "util/utils.h"


// state
static std::vector<SETUPAPI_SETTINGS> SETUPAPI_SETTINGS_LIST;
static uint8_t GUID_ZERO[16]{};


// SETUPAPI hook types
typedef HDEVINFO (WINAPI *SetupDiGetClassDevsA_t)(const GUID *ClassGuid, LPCSTR Enumerator, HWND hwndParent,
                                                  DWORD Flags);

static SetupDiGetClassDevsA_t SetupDiGetClassDevsA_real = nullptr;

typedef HDEVINFO (WINAPI *SetupDiGetClassDevsW_t)(const GUID *ClassGuid, LPCWSTR Enumerator, HWND hwndParent,
                                                  DWORD Flags);

static SetupDiGetClassDevsW_t SetupDiGetClassDevsW_real = nullptr;

typedef BOOL (WINAPI *SetupDiEnumDeviceInfo_t)(HDEVINFO DeviceInfoSet, DWORD MemberIndex,
                                               PSP_DEVINFO_DATA DeviceInfoData);

static SetupDiEnumDeviceInfo_t SetupDiEnumDeviceInfo_real = nullptr;

typedef BOOL (WINAPI *SetupDiEnumDeviceInterfaces_t)(HDEVINFO DeviceInfoSet, PSP_DEVINFO_DATA DeviceInfoData,
                                                     const GUID *InterfaceClassGuid, DWORD MemberIndex,
                                                     PSP_DEVICE_INTERFACE_DATA DeviceInterfaceData);

static SetupDiEnumDeviceInterfaces_t SetupDiEnumDeviceInterfaces_real = nullptr;

typedef BOOL (WINAPI *SetupDiGetDeviceRegistryPropertyA_t)(HDEVINFO DeviceInfoSet, PSP_DEVINFO_DATA DeviceInfoData,
                                                           DWORD Property, PDWORD PropertyRegDataType,
                                                           PBYTE PropertyBuffer, DWORD PropertyBufferSize,
                                                           PDWORD RequiredSize);

static SetupDiGetDeviceRegistryPropertyA_t SetupDiGetDeviceRegistryPropertyA_real = nullptr;

typedef BOOL (WINAPI *SetupDiGetDeviceInterfaceDetailA_t)(HDEVINFO DeviceInfoSet,
                                                          PSP_DEVICE_INTERFACE_DATA DeviceInterfaceData,
                                                          PSP_DEVICE_INTERFACE_DETAIL_DATA DeviceInterfaceDetailData,
                                                          DWORD DeviceInterfaceDetailDataSize, PDWORD RequiredSize,
                                                          PSP_DEVINFO_DATA DeviceInfoData);

static SetupDiGetDeviceInterfaceDetailA_t SetupDiGetDeviceInterfaceDetailA_real = nullptr;

typedef BOOL (WINAPI *SetupDiGetDeviceInterfaceDetailW_t)(HDEVINFO DeviceInfoSet,
                                                          PSP_DEVICE_INTERFACE_DATA DeviceInterfaceData,
                                                          PSP_DEVICE_INTERFACE_DETAIL_DATA DeviceInterfaceDetailData,
                                                          DWORD DeviceInterfaceDetailDataSize, PDWORD RequiredSize,
                                                          PSP_DEVINFO_DATA DeviceInfoData);

static SetupDiGetDeviceInterfaceDetailW_t SetupDiGetDeviceInterfaceDetailW_real = nullptr;

typedef BOOL (WINAPI *SetupDiDestroyDeviceInfoList_t)(HDEVINFO DeviceInfoSet);

static SetupDiDestroyDeviceInfoList_t SetupDiDestroyDeviceInfoList_real = nullptr;

typedef BOOL (WINAPI *SetupDiGetDeviceInfoListDetailA_t)(HDEVINFO, PSP_DEVINFO_LIST_DETAIL_DATA);

static SetupDiGetDeviceInfoListDetailA_t SetupDiGetDeviceInfoListDetailA_real = nullptr;


static WINAPI HDEVINFO SetupDiGetClassDevsA_hook(const GUID *ClassGuid, LPCSTR Enumerator, HWND hwndParent,
                                                 DWORD Flags) {

    // custom
    for (size_t i = 0; i < SETUPAPI_SETTINGS_LIST.size(); i++) {
        SETUPAPI_SETTINGS *settings = &SETUPAPI_SETTINGS_LIST[i];
        void* guid = ClassGuid ? (void*) ClassGuid : GUID_ZERO;
        if (!memcmp(guid, settings->class_guid, 16)) {
            SetLastError(ERROR_SUCCESS);
            return (HDEVINFO) &SETUPAPI_SETTINGS_LIST[i];
        }
    }

    // fallback
    return SetupDiGetClassDevsA_real(ClassGuid, Enumerator, hwndParent, Flags);
}

static WINAPI HDEVINFO SetupDiGetClassDevsW_hook(const GUID *ClassGuid, LPCWSTR Enumerator, HWND hwndParent,
                                                 DWORD Flags) {

    // custom
    for (size_t i = 0; i < SETUPAPI_SETTINGS_LIST.size(); i++) {
        SETUPAPI_SETTINGS *settings = &SETUPAPI_SETTINGS_LIST[i];
        if (ClassGuid && !memcmp(ClassGuid, settings->class_guid, 16)) {
            SetLastError(ERROR_SUCCESS);
            return (HDEVINFO) &SETUPAPI_SETTINGS_LIST[i];
        }
    }

    // fallback
    return SetupDiGetClassDevsW_real(ClassGuid, Enumerator, hwndParent, Flags);
}

static WINAPI BOOL SetupDiEnumDeviceInfo_hook(HDEVINFO DeviceInfoSet, DWORD MemberIndex,
                                              PSP_DEVINFO_DATA DeviceInfoData) {

    // custom
    for (size_t i = 0; i < SETUPAPI_SETTINGS_LIST.size(); i++) {
        if (DeviceInfoSet == &SETUPAPI_SETTINGS_LIST[i]) {
            if (MemberIndex) {
                SetLastError(ERROR_NO_MORE_ITEMS);
                return false;
            } else {
                SetLastError(ERROR_SUCCESS);
                return true;
            }
        }
    }

    // fallback
    return SetupDiEnumDeviceInfo_real(DeviceInfoSet, MemberIndex, DeviceInfoData);
}

static WINAPI BOOL SetupDiEnumDeviceInterfaces_hook(HDEVINFO DeviceInfoSet, PSP_DEVINFO_DATA DeviceInfoData,
                                                    const GUID *InterfaceClassGuid, DWORD MemberIndex,
                                                    PSP_DEVICE_INTERFACE_DATA DeviceInterfaceData) {
    // custom
    for (size_t i = 0; i < SETUPAPI_SETTINGS_LIST.size(); i++) {
        if (DeviceInfoSet == &SETUPAPI_SETTINGS_LIST[i]) {
            SetLastError(ERROR_SUCCESS);
            return MemberIndex == 0;
        }
    }

    // fallback
    return SetupDiEnumDeviceInterfaces_real(DeviceInfoSet, DeviceInfoData, InterfaceClassGuid, MemberIndex,
                                            DeviceInterfaceData);
}

static WINAPI BOOL SetupDiGetDeviceRegistryPropertyA_hook(HDEVINFO DeviceInfoSet, PSP_DEVINFO_DATA DeviceInfoData,
                                                          DWORD Property, PDWORD PropertyRegDataType,
                                                          PBYTE PropertyBuffer, DWORD PropertyBufferSize,
                                                          PDWORD RequiredSize) {

    // custom
    for (size_t i = 0; i < SETUPAPI_SETTINGS_LIST.size(); i++) {
        SETUPAPI_SETTINGS *settings = &SETUPAPI_SETTINGS_LIST[i];
        if (DeviceInfoSet == settings) {

            // dummy data
            void* property_data = nullptr;
            DWORD property_size = 0;

            // get property data
            switch (Property) {
                case SPDRP_DEVICEDESC:
                    property_data = settings->property_devicedesc;
                    property_size = (DWORD) (strlen(settings->property_devicedesc) + 1);
                    break;
                case SPDRP_HARDWAREID:
                    property_data = settings->property_hardwareid;
                    property_size = (DWORD) (strlen(settings->property_hardwareid) + 1);
                    break;
                case SPDRP_ADDRESS:
                    /*
                     * The property index thing is just because IIDX 25+ requires two different addresses to
                     * be spoofed for cameras. Doing this in a clean way requires a complete hook rewrite.
                     */
                    static int property_index = 0;
                    property_data = &settings->property_address[property_index];
                    property_size = sizeof(DWORD);
                    property_index = (property_index + 1) & 1;
                    break;
                default:
                    log_warning("setupapi", "Unknown property type: " + to_string(Property));
                    SetLastError(ERROR_INVALID_DATA);
                    return false;
            }

            // check buffer size
            if (PropertyBufferSize < property_size) {
                if (RequiredSize)
                    *RequiredSize = property_size;
                SetLastError(ERROR_INSUFFICIENT_BUFFER);
                return false;
            }

            // write property
            if (PropertyRegDataType)
                *PropertyRegDataType = 1;
            memcpy(PropertyBuffer, property_data, property_size);
            SetLastError(ERROR_SUCCESS);
            return true;
        }
    }

    // fallback
    return SetupDiGetDeviceRegistryPropertyA_real(DeviceInfoSet, DeviceInfoData, Property, PropertyRegDataType,
                                                  PropertyBuffer, PropertyBufferSize, RequiredSize);
}

static WINAPI BOOL SetupDiGetDeviceInterfaceDetailA_hook(HDEVINFO DeviceInfoSet,
                                                         PSP_DEVICE_INTERFACE_DATA DeviceInterfaceData,
                                                         PSP_DEVICE_INTERFACE_DETAIL_DATA DeviceInterfaceDetailData,
                                                         DWORD DeviceInterfaceDetailDataSize, PDWORD RequiredSize,
                                                         PSP_DEVINFO_DATA DeviceInfoData) {
    // custom
    for (size_t i = 0; i < SETUPAPI_SETTINGS_LIST.size(); i++) {
        SETUPAPI_SETTINGS *settings = &SETUPAPI_SETTINGS_LIST[i];
        if (DeviceInfoSet == settings) {

            // get data length
            DWORD data_size = (DWORD) (strlen(settings->interface_detail) + 1);

            // check size
            if (DeviceInterfaceDetailDataSize < data_size) {
                SetLastError(ERROR_INSUFFICIENT_BUFFER);
                *RequiredSize = data_size;
                return false;
            }

            // clear error
            SetLastError(ERROR_SUCCESS);

            // copy data size + data
            *((DWORD *) DeviceInterfaceDetailData) = data_size;
            memcpy(DeviceInterfaceDetailData->DevicePath, settings->interface_detail, data_size);

            // success
            return true;
        }
    }

    // fallback
    return SetupDiGetDeviceInterfaceDetailA_real(DeviceInfoSet, DeviceInterfaceData, DeviceInterfaceDetailData,
                                                 DeviceInterfaceDetailDataSize, RequiredSize, DeviceInfoData);
}

static WINAPI BOOL SetupDiGetDeviceInterfaceDetailW_hook(HDEVINFO DeviceInfoSet,
                                                         PSP_DEVICE_INTERFACE_DATA DeviceInterfaceData,
                                                         PSP_DEVICE_INTERFACE_DETAIL_DATA DeviceInterfaceDetailData,
                                                         DWORD DeviceInterfaceDetailDataSize, PDWORD RequiredSize,
                                                         PSP_DEVINFO_DATA DeviceInfoData) {
    // custom
    for (size_t i = 0; i < SETUPAPI_SETTINGS_LIST.size(); i++) {
        SETUPAPI_SETTINGS *settings = &SETUPAPI_SETTINGS_LIST[i];
        if (DeviceInfoSet == settings) {

            // convert to wstring
            std::wstring ws = s2ws(std::string(settings->interface_detail));

            // get data length
            DWORD data_size = (DWORD) (ws.length() * 2 + 2);
            DWORD struct_size = data_size + sizeof(DWORD);

            // check size
            if (DeviceInterfaceDetailDataSize < struct_size) {
                SetLastError(ERROR_INSUFFICIENT_BUFFER);
                *RequiredSize = struct_size;
                return false;
            }

            // clear error
            SetLastError(ERROR_SUCCESS);

            // copy data size + data
            DeviceInterfaceDetailData->cbSize = struct_size;
            memcpy(DeviceInterfaceDetailData->DevicePath, ws.c_str(), data_size);

            // success
            return true;
        }
    }

    // fallback
    return SetupDiGetDeviceInterfaceDetailW_real(DeviceInfoSet, DeviceInterfaceData, DeviceInterfaceDetailData,
                                                 DeviceInterfaceDetailDataSize, RequiredSize, DeviceInfoData);
}

static WINAPI BOOL SetupDiDestroyDeviceInfoList_hook(HDEVINFO DeviceInfoSet) {

    // custom
    for (size_t i = 0; i < SETUPAPI_SETTINGS_LIST.size(); i++) {
        if (DeviceInfoSet == &SETUPAPI_SETTINGS_LIST[i])
            return true;
    }

    // fallback
    return SetupDiDestroyDeviceInfoList_real(DeviceInfoSet);
}

static WINAPI BOOL SetupDiGetDeviceInfoListDetailA_hook(HDEVINFO DeviceInfoSet,
                                                        PSP_DEVINFO_LIST_DETAIL_DATA DeviceInfoSetDetailData) {
    // custom
    for (size_t i = 0; i < SETUPAPI_SETTINGS_LIST.size(); i++) {
        if (DeviceInfoSet == &SETUPAPI_SETTINGS_LIST[i])
            return true;
    }

    // fallback
    return SetupDiGetDeviceInfoListDetailA_real(DeviceInfoSet, DeviceInfoSetDetailData);
}

void setupapihook_init(HINSTANCE module) {

    // check for skip
    if (!detour::iat_find("SetupDiGetClassDevsA", module) && !detour::iat_find("SetupDiGetClassDevsW", module)) {
        log_info("setupapi", "no usage detected in module");
        return;
    }

    // hook functions
    SetupDiGetClassDevsA_real = (SetupDiGetClassDevsA_t) detour::iat_try(
                    "SetupDiGetClassDevsA", (void *) &SetupDiGetClassDevsA_hook, module);
    SetupDiGetClassDevsW_real = (SetupDiGetClassDevsW_t) detour::iat_try(
                    "SetupDiGetClassDevsW", (void *) &SetupDiGetClassDevsW_hook, module);
    SetupDiEnumDeviceInfo_real = (SetupDiEnumDeviceInfo_t) detour::iat_try(
                    "SetupDiEnumDeviceInfo", (void *) &SetupDiEnumDeviceInfo_hook, module);
    SetupDiEnumDeviceInterfaces_real = (SetupDiEnumDeviceInterfaces_t) detour::iat_try(
                    "SetupDiEnumDeviceInterfaces", (void *) &SetupDiEnumDeviceInterfaces_hook, module);
    SetupDiGetDeviceRegistryPropertyA_real = (SetupDiGetDeviceRegistryPropertyA_t) detour::iat_try(
                    "SetupDiGetDeviceRegistryPropertyA",
                    (void *) &SetupDiGetDeviceRegistryPropertyA_hook, module);
    SetupDiGetDeviceInterfaceDetailA_real = (SetupDiGetDeviceInterfaceDetailA_t) detour::iat_try(
                    "SetupDiGetDeviceInterfaceDetailA",
                    (void *) &SetupDiGetDeviceInterfaceDetailA_hook, module);
    SetupDiGetDeviceInterfaceDetailW_real = (SetupDiGetDeviceInterfaceDetailW_t) detour::iat_try(
                    "SetupDiGetDeviceInterfaceDetailW",
                    (void *) &SetupDiGetDeviceInterfaceDetailW_hook, module);
    SetupDiDestroyDeviceInfoList_real = (SetupDiDestroyDeviceInfoList_t) detour::iat_try(
                    "SetupDiDestroyDeviceInfoList", (void *) &SetupDiDestroyDeviceInfoList_hook, module);
    SetupDiGetDeviceInfoListDetailA_real = (SetupDiGetDeviceInfoListDetailA_t) detour::iat_try(
                    "SetupDiGetDeviceInfoListDetailA", (void *) &SetupDiGetDeviceInfoListDetailA_hook, module);
}

void setupapihook_add(SETUPAPI_SETTINGS settings) {
    SETUPAPI_SETTINGS_LIST.push_back(settings);
}
